local ElitistGroup = select(2, ...)
local L = ElitistGroup.L

function ElitistGroup:ADDON_LOADED(event, addon)
	if( addon ~= "ElitistGroupSender" ) then return end
	self.frame:UnregisterEvent("ADDON_LOADED")
	
	ElitistGroupSenderDB = ElitistGroupSenderDB or {
		channels = {BATTLEGROUND = false, PARTY = true, RAID = true, GUILD = true, WHISPER = true},
	}
	
	self.db = ElitistGroupSenderDB
	self.playerID = string.format("%s-%s", UnitName("player"), GetRealmName())
end

-- Save the mains experience data
function ElitistGroup:PLAYER_LOGOUT()
	if( self.db.main ~= self.playerID ) then return end
	local mainData = {}
	
	for achievementID in pairs(self.VALID_ACHIEVEMENTS) do
		local id, _, _, completed, _, _, _, _, flags = GetAchievementInfo(achievementID)
		if( bit.band(flags, ACHIEVEMENT_FLAGS_STATISTIC) > 0 ) then
			local statistic = tonumber(GetStatistic(id))
			if( statistic ) then
				table.insert(mainData, id)
				table.insert(mainData, statistic)
			end
		elseif( completed ) then
			table.insert(mainData, id)
			table.insert(mainData, 1)
		end
	end
	
	self.db.mainData = #(mainData) > 0 and table.concat(mainData, "@") or nil
end

-- Fun part, figure out everything to save
local cachedPlayerData, compressedPlayerData, playerData

local function writeTable(tbl)
	local data = ""
	for key, value in pairs(tbl) do
		local valueType = type(value)

		-- Wrap the key in brackets if it's a number
		if( type(key) == "number" ) then
			key = string.format("[%s]", key)
		-- This will match any punctuation, spacing or control characters, basically anything that requires wrapping around them
		elseif( string.match(key, "[%p%s%c]") ) then
			key = string.format("[\"%s\"]", key)
		end

		-- foo = {bar = 5}
		if( valueType == "table" ) then
			data = string.format("%s%s=%s;", data, key, writeTable(value))
		-- foo = true / foo = 5
		elseif( valueType == "number" or valueType == "boolean" ) then
			data = string.format("%s%s=%s;", data, key, tostring(value))
		-- foo = "bar"
		else
			data = string.format("%s%s=\"%s\";", data, key, value)
		end
	end

	return "{" .. data .. "}"
end

local function getTalentData(classToken, activeTalentGroup)
	local specRole
	local forceData = ElitistGroup.SPEC_OVERRIDE[classToken]
	if( forceData ) then
		local talentMatches = 0
		for tabIndex=1, GetNumTalentTabs() do
			for talentID=1, GetNumTalents(tabIndex) do
				local name, _, _, _, spent = GetTalentInfo(tabIndex, talentID, nil, nil, activeTalentGroup)
				if( forceData[name] and spent >= forceData[name] ) then
					talentMatches = talentMatches + 1
				end
			end
		end

		specRole = talentMatches >= forceData.required and forceData.role or nil
	end

	local first = select(3, GetTalentTabInfo(1, nil, nil, activeTalentGroup)) or 0
	local second = select(3, GetTalentTabInfo(2, nil, nil, activeTalentGroup)) or 0
	local third = select(3, GetTalentTabInfo(3, ni, nil, activeTalentGroup)) or 0
	local unspentPoints = GetUnspentTalentPoints(nil, nil, activeTalentGroup)
	unspentPoints = unspentPoints > 0 and unspentPoints or nil
	
	return first, second, third, specRole, unspentPoints
end

local EQUIP_SLOTS = {["HeadSlot"] = true, ["ChestSlot"] = true, ["RangedSlot"] = true, ["WristSlot"] = true, ["Trinket1Slot"] = true, ["Trinket0Slot"] = true, ["MainHandSlot"] = true, ["SecondaryHandSlot"] = true, ["Finger0Slot"] = true, ["Finger1Slot"] = true, ["NeckSlot"] = true, ["FeetSlot"] = true, ["LegsSlot"] = true, ["WaistSlot"] = true, ["HandsSlot"] = true, ["BackSlot"] = true, ["ShoulderSlot"] = true}
function ElitistGroup:GetPlayerData()
	if( not playerData ) then
		local name, server = UnitName("player")
		playerData = {talentTree1 = 0, talentTree2 = 0, talentTree3 = 0, achievements = {}, equipment = {}}
		playerData.name = name
		playerData.server = server and server ~= "" and server or GetRealmName()
		playerData.level = UnitLevel("player")
		playerData.classToken = select(2, UnitClass("player"))
	else
		table.wipe(playerData.equipment)
		table.wipe(playerData.achievements)
	end
	
	-- Save equipment
	for equipType in pairs(EQUIP_SLOTS) do
		local inventoryID = GetInventorySlotInfo(equipType)
		local itemLink = GetInventoryItemLink("player", inventoryID)
		if( itemLink ) then
			playerData.equipment[inventoryID] = string.match(itemLink, "|H(.-)|h")
		end
	end
	
	-- Save talents
	local activeTalentGroup = GetActiveTalentGroup()
	local first, second, third, specRole, unspentPoints = getTalentData(playerData.classToken, activeTalentGroup)
	playerData.talentTree1 = first
	playerData.talentTree2 = second
	playerData.talentTree3 = third
	playerData.specRole = specRole
	playerData.unspentPoints = unspentPoints
	
	if( GetNumTalentGroups() > 1 ) then
		local first, second, third, specRole, unspentPoints = getTalentData(playerData.classToken, activeTalentGroup == 2 and 1 or 2)
		playerData.secondarySpec = playerData.secondarySpec or {}
		playerData.secondarySpec.talentTree1 = first
		playerData.secondarySpec.talentTree2 = second
		playerData.secondarySpec.talentTree3 = third
		playerData.secondarySpec.specRole = specRole
		playerData.secondarySpec.unspentPoints = unspentPoints
	end
	
	-- Save achievements
	for achievementID in pairs(self.VALID_ACHIEVEMENTS) do
		local id, _, _, completed, _, _, _, _, flags = GetAchievementInfo(achievementID)
		if( bit.band(flags, ACHIEVEMENT_FLAGS_STATISTIC) > 0 ) then
			playerData.achievements[id] = tonumber(GetStatistic(id)) or nil
		else
			playerData.achievements[id] = completed and 1 or nil
		end
	end

	-- Finally, convert it all into the serialized form for us to save and send
	return writeTable(playerData)
end

-- Send gear + main experience
local AceComm = LibStub("AceComm-3.0")
function ElitistGroup:SendGearData(sender)
	-- Players info should rarely change, so we can just cache it and that will be all we need most of the time
	if( not cachedPlayerData ) then
		if( self.db.mainData and self.db.main ~= self.playerID ) then
			cachedPlayerData = string.format("GEAR@%s@%s", self:GetPlayerData(), self.db.mainData)
		else
			cachedPlayerData = string.format("GEAR@%s", self:GetPlayerData())
		end
	end
	
	AceComm:SendCommMessage("ELITG", cachedPlayerData, "WHISPER", sender)
end

-- Send the players main experience data only
function ElitistGroup:SendMainData(sender)
	if( self.db.main == self.playerID or not self.db.mainData ) then return end
	AceComm:SendCommMessage("ELITG", string.format("MAIN@%s", self.db.mainData), "WHISPER", sender)
end

-- Don't need to use AceComm-3.0 except for the message splitting it provides, because all messages we handle are <20 characters
local requestThrottle, combatQueue = {}, {}
local playerName = UnitName("player")
function ElitistGroup:CHAT_MSG_ADDON(event, prefix, message, distribution, sender)
	if( prefix ~= "ELITG" or sender == playerName or not self.db.channels[distribution] or ( requestThrottle[sender] and requestThrottle[sender] > GetTime() ) ) then return end
	requestThrottle[sender] = GetTime() + (message == "REQGEAR" and 60 or 3600)

	if( InCombatLockdown() ) then
		self.frame:RegisterEvent("PLAYER_REGEN_ENABLED")
		
		-- If somebody is requesting gear and we queued main, will allow the swap because gear is main data nayway
		if( not combatQueue[sender] or combatQueue[sender] == "REQMAIN" ) then
			combatQueue[sender] = message
		end
		return
	end
	
	if( message == "REQGEAR" ) then
		self:SendGearData(sender)
	elseif( message == "REQMAIN" ) then
		self:SendMainData(sender)
	end
end

-- Whenever the equipment or achievements change, will want to reset cache
function ElitistGroup:ResetPlayerCache()
	cachedPlayerData = nil
end

ElitistGroup.PLAYER_EQUIPMENT_CHANGED = ElitistGroup.ResetPlayerCache
ElitistGroup.ACHIEVEMENT_EARNED = ElitistGroup.ResetPlayerCache
ElitistGroup.ACTIVE_TALENT_GROUP_CHANGED = ElitistGroup.ResetPlayerCache

-- Combat queuer of sending messages
function ElitistGroup:PLAYER_REGEN_ENABLED()
	self.frame:UnregisterEvent("PLAYER_REGEN_ENABLED")
	
	for sender, message in pairs(combatQueue) do
		self:CHAT_MSG_ADDON(nil, "ELITG", message, "WHISPER", sender)
	end
	
	table.wipe(combatQueue)
end

-- Throttle manager
function ElitistGroup:PLAYER_LEAVING_WORLD()
	local time = GetTime()
	for name, endTime in pairs(requestThrottle) do
		if( endTime <= time ) then
			requestThrottle[name] = nil
		end
	end
end

-- Event handler
local frame = CreateFrame("Frame")
frame:RegisterEvent("ADDON_LOADED")
frame:RegisterEvent("CHAT_MSG_ADDON")
frame:RegisterEvent("PLAYER_LOGOUT")
frame:RegisterEvent("PLAYER_EQUIPMENT_CHANGED")
frame:RegisterEvent("ACHIEVEMENT_EARNED")
frame:RegisterEvent("PLAYER_LEAVING_WORLD")
frame:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")
frame:SetScript("OnEvent", function(self, event, ...)
	ElitistGroup[event](ElitistGroup, event, ...)
end)

ElitistGroup.frame = frame

-- Slash command
function ElitistGroup:Print(msg)
	DEFAULT_CHAT_FRAME:AddMessage("|cff33ff99Elitist Group|r: " .. msg)
end

SLASH_EGSENDER1 = "/egs"
SLASH_EGSENDER2 = "/elitistgroupsender"
SLASH_EGSENDER3 = "/elitistgroupsend"
SLASH_EGSENDER4 = "/egsend"
SlashCmdList["EGSENDER"] = function(msg)
	msg = string.lower(msg or "")
	
	if( msg == "main" ) then
		ElitistGroup:ResetPlayerCache()
		ElitistGroup.db.main = ElitistGroup.db.main ~= ElitistGroup.playerID and ElitistGroup.playerID or nil
		ElitistGroup.db.mainData = nil
		
		if( ElitistGroup.db.main ) then
			ElitistGroup:Print(string.format(L["Set %s as your main."], ElitistGroup.playerID))
		else
			ElitistGroup:Print(string.format(L["No longer using %s as your main."], ElitistGroup.playerID))
		end
		
	elseif( string.match(msg, "send (.+)") ) then
		local name = string.match(msg, "send (.+)")
		if( not name or name == "" ) then
			ElitistGroupt:Print(L["No player name entered to send gear data to."])
			return
		end
		
		ElitistGroup:Print(string.format(L["Sent your gear to %s, may take a minute or two to show up for them."], name))
		ElitistGroup:SendGearData(name)
		
	elseif( msg == "list" ) then
		ElitistGroup:Print(L["Listing configuration"])
		if( ElitistGroup.db.main ) then
			DEFAULT_CHAT_FRAME:AddMessage(string.format(L["%s is set as your main."], ElitistGroup.db.main))
		else
			DEFAULT_CHAT_FRAME:AddMessage(L["No characters set as your main yet."])
		end
		
		local channels = {}
		table.insert(channels, ElitistGroup.db.channels.RAID and L["raid"] or nil)
		table.insert(channels, ElitistGroup.db.channels.PARTY and L["party"] or nil)
		table.insert(channels, ElitistGroup.db.channels.GUILD and L["guild"] or nil)
		table.insert(channels, ElitistGroup.db.channels.BATTLEGROUND and L["battleground"] or nil)
			
		if( #(channels) == 0 ) then
			DEFAULT_CHAT_FRAME:AddMessage(L["Only watching addon communications over the whisper channel."])
		else
			DEFAULT_CHAT_FRAME:AddMessage(string.format(L["Watching for addon communications in: %s"], table.concat(channels, ", ")))
		end
		
	elseif( string.match(msg, "channel (.+)") ) then
		local channel = string.upper(string.match(msg, "channel (.+)") or "")
		if( channel ~= "PARTY" and channel ~= "GUILD" and channel ~= "RAID" and channel ~= "BATTLEGROUND" ) then
			ElitistGroup:Print(string.format(L["\"%s\" is an invalid channel to toggle comms in."], channel))
			return
		end
		
		ElitistGroup.db.channels[channel] = not ElitistGroup.db.channels[channel]
		
		if( ElitistGroup.db.channels[channel] ) then
			ElitistGroup:Print(string.format(L["Now responding to addon communications in %s."], string.lower(channel)))
		else
			ElitistGroup:Print(string.format(L["No longer responding to addon communications in %s."], string.lower(channel)))
		end
	else
		ElitistGroup:Print(L["Slash commands"])
		DEFAULT_CHAT_FRAME:AddMessage(L["/egs channel <raid/party/battleground/guild> - Toggles the addon communication channels to respond to"])
		DEFAULT_CHAT_FRAME:AddMessage(L["/egs main - Sets this character as your main, to pull experience data from"])
		DEFAULT_CHAT_FRAME:AddMessage(L["/egs send <name> - Sends your gear to <name> so they do not have to request it"])
		DEFAULT_CHAT_FRAME:AddMessage(L["/egs list - Lists your current settings"])
	end
end



